@@ -11,7 +11,10 @@ import xxhash
 import zarr
 from dask import delayed
 from dask.distributed import as_completed
+from prefect import Flow, Parameter, task, unmapped
+from prefect.engine.executors import DaskExecutor
 
+from utoolbox.pipeline.tasks import CalcXxHash, ZarrWriteArray
 from utoolbox.util.dask import get_client
 
 from ..base import (
@@ -168,8 +171,10 @@ class ZarrDataset(
 
         hash_gen = xxhash.xxh64()
 
-        client = client if client else get_client()
-        futures = {}  # conversion tasks, batch submit after populated all of them
+        # client = client if client else get_client()
+        # futures = {}  # conversion tasks, batch submit after populated all of them
+
+        tasks = []
 
         # start populating the container structure
         #   /time/channel/setup/level
@@ -241,6 +246,8 @@ class ZarrDataset(
                         if label in s_root:
                             label_group = s_root[label]
 
+                            # TODO start prefect task from here
+
                             # get the hash
                             hash_gen.update(data.compute())
                             src_hash = hash_gen.hexdigest()
@@ -325,40 +332,86 @@ class ZarrDataset(
                         # NOTE using default chunk shape after rechunk will cause
                         # problem, since chunk size is composite of chunks as tuples
                         # instead of int
-                        data_src = data.rechunk(data_dst.chunks)
-                        array = data_src.to_zarr(
-                            data_dst,
-                            overwrite=overwrite,
-                            compute=False,
-                            return_stored=True,
-                        )
+                        # [A]
+                        # data_src = data.rechunk(data_dst.chunks)
+                        # array = data_src.to_zarr(
+                        #    data_dst, overwrite=True, compute=False, return_stored=True,
+                        # )
+                        # [B]
+                        tasks.append(
+                            (label_group, data, data_dst)
+                        )  # append new request
 
-                        @delayed
-                        def calc_checksum(array):
-                            # NOTE cannot create external dependency in delayed funcs,
-                            # map futures to their target groups instead
-                            return xxhash.xxh64(array).hexdigest()
+                        # [A]
+                        # @delayed
+                        # def calc_checksum(array):
+                        #    # NOTE cannot create external dependency in delayed funcs,
+                        #    # map futures to their target groups instead
+                        #    return xxhash.xxh64(array).hexdigest()
 
-                        checksum = calc_checksum(array)
-                        future = client.compute(checksum)
-                        futures[future] = label_group
+                        # [A]
+                        # checksum = calc_checksum(array)
+                        # future = client.compute(checksum)
+                        # futures[future] = label_group
                         # TODO add callback here to update progressbar
 
-        if not futures:
+        # [A]
+        # if not futures:
+        #    return  # nothing to do
+        # [B]
+        if not tasks:
             return  # nothing to do
 
-        n_failed = 0
-        for future in as_completed(futures.keys()):
-            try:
-                group, checksum = futures[future], future.result()
-            except Exception:
-                logger.error(f'failed to serialize "{group.path}"')
-                n_failed += 1
-            else:
-                logger.debug(f'"{group.path}" xxh64="{checksum}"')
-                group.attrs["checksum"] = checksum
-        if n_failed > 0:
-            logger.error(f"{n_failed} failed serialization task(s)")
+        # [B]
+        write_array = ZarrWriteArray(rechunk="auto")
+        calc_checksum = CalcXxHash(method="xxh64", return_format="hex")
+
+        @task
+        def extract_tuple(pairs, index):
+            return pairs[index]
+
+        @task
+        def write_checksum(group, checksum):
+            logger.debug(f'"{group.path}" xxh64="{checksum}"')
+            group.attrs["checksum"] = checksum
+
+        with Flow("Dump ZarrDataset") as flow:
+            # build work flow
+
+            pairs = Parameter("tasks")
+
+            group = extract_tuple.map(pairs, unmapped(0))
+            src = extract_tuple.map(pairs, unmapped(1))
+            dst = extract_tuple.map(pairs, unmapped(2))
+
+            dst_zarr = write_array.map(
+                src,
+                dst,
+                overwrite=unmapped(True),
+                compute=unmapped(False),
+                return_stored=unmapped(True),
+            )
+
+            checksum = calc_checksum.map(dst_zarr)
+
+            write_checksum.map(group, checksum)
+
+        executor = DaskExecutor(address="tcp://localhost:8786")
+        flow.run(tasks=tasks, executor=executor)
+
+        # [A]
+        # n_failed = 0
+        # for future in as_completed(futures.keys()):
+        #    try:
+        #        group, checksum = futures[future], future.result()
+        #    except Exception:
+        #        logger.error(f'failed to serialize "{group.path}"')
+        #        n_failed += 1
+        #    else:
+        #        logger.debug(f'"{group.path}" xxh64="{checksum}"')
+        #        group.attrs["checksum"] = checksum
+        # if n_failed > 0:
+        #    logger.error(f"{n_failed} failed serialization task(s)")
 
     ##
 
@@ -679,11 +732,42 @@ class ZarrDataset(
 
 
 class MutableZarrDataset(ZarrDataset):
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.active_label = None
+
+    def __setitem__(self, key, value):
+        # TODO key -> uuid reference, value -> data to write in the corresponding group
+        pass
+
+    ##
+
+    @property
+    def active_label(self) -> str:
+        """Active label are the one we are currently mutating."""
+        if self._active_label is None:
+            self.active_label = self.label  # NOTE should trigger another warning
+        return self._active_label
+
+    @active_label.setter
+    def active_label(self, label: Optional[str]) -> str:
+        if label == self.label:
+            logger.warning(
+                f'mutate on inventoried label "{label}" will lead to data inconsistency, please reload after modification'
+            )
+        self._active_label = label
+
+    ##
+
     @classmethod
     def from_immutable(cls, dataset: ZarrDataset):
         """From immutable Zarr dataset."""
-        dataset.__class__ = cls  # inherited, should be fine using this hacky method
-        return dataset
+        name = os.path.basename(dataset.root_dir)
+        logger.info(f'reloading "{name}" as mutable Zarr dataset')
+
+        # constructor needs:
+        #   store / label / level / path
+        return cls(dataset.root_dir, dataset.label, dataset.level, dataset.path)
 
     ##
 
